// TENEOSTD
//////////////////////////////////////////////////////////////////////
// haptics.cpp: implementation of the haptics class.
//
// NOTE:  This is applicaiton specific code for HDAL based haptics.
//
//
// haptics class provides the interface between the application and
// the haptic device via HDAL.
//
// The haptic device is treated as the world coordinate system.
//
// - Each servo tick the callback ContactCB is called.
//
// - ContactCB queries end-effector characteristics (pos, transform, button).
//   It then calls the calculateForce object associated with the tObject sending
//   the resultant calculation to the haptic device.
//
// - Client thread access to the haptics class should be proceeded by a
//   calling haptics::syncFromServo().  The call syncs the client thread
//   copy of the end-effector characteristics from the servo thread using
//   GetStateCB.
//
//////////////////////////////////////////////////////////////////////

#include "../stdafx.h"

#include "haptics.h"

#include <math.h>
#include <assert.h>
#include <direct.h>

#include "orientation.h"
#include "matrix.h"


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
//#define new DEBUG_NEW
#endif


// Constants used to tune haptic feel
// These are "master" defaults, which will be in effect if the
//   HDAL device type is unknown.  See initParams() for known
//   device type defaults.
static const char*  g_deviceType;

const double SCALE_FACTOR = 1000;

haptics* haptics::s_instance = 0;

static HDLDeviceHandle g_hDevice = HDL_INVALID_HANDLE;
static bool g_isServoDone = false;
static DWORD g_servoThreadID;
static HANDLE g_hServoThreadHandle;


//#define LOG

#ifdef LOG
FILE* fd = 0;
#endif


//-------------------------------------------------------
// tuning vars from the .INI file
//-------------------------------------------------------

//---- general purpose vars (don't need to be tuned for each haptics device)----

// camera values
float SANDBOX_CAMERA_DIST_BEHIND =			-1500.0f;	// distance from car to place sandbox camera when returning from driving
float SANDBOX_CAMERA_DIST_ABOVE =			800.0f;

float CAR_CAMERA_1ST_DIST_BEHIND =			-50.0f;		// distance from car to place driving camera (1st person camera)
float CAR_CAMERA_1ST_DIST_ABOVE =			25.0f;

float CAR_CAMERA_3RD_DIST_BEHIND =			-800.0f;	// distance from car to place driving camera (3rd person camera)
float CAR_CAMERA_3RD_DIST_ABOVE =			300.0f;

float CAMERA_MOVE_TIME =					25.0f;		// time the camera sandbox/driving transition takes

// general game physics values
float PHYSICS_GRAVITY =						1800.0f;	// gravity

// generic physics object values
float OBJ_HOLDING_VELOCITY_SCALE =			10.0f;		// velocity scale to use when holding and dragging the object

// flowerpot physics object values
float FLOWERPOT_PHYSICS_MASS =				150.0f;		// mass of the object
float FLOWERPOT_PHYSICS_FRICTION =			400.0f;		// friction applied to the object
float FLOWERPOT_PHYSICS_DAMPING =			50.0f;		// amount of damping

float FLOWERPOT_SCALE =						1.0f;		// display scale
float FLOWERPOT_COLLIDE_PICKUP_SIZE =		150.0f;		// pickup size of the object
float FLOWERPOT_COLLIDE_NEAR_SIZE =			300.0f;		// near (run haptics) size of the object
float FLOWERPOT_HAPTICS_RADIUS =			30.0f;		// radius of the haptics sphere

// crate physics object values
float CRATE_PHYSICS_MASS =					70.0f;		// mass of the object
float CRATE_PHYSICS_FRICTION =				400.0f;		// friction applied to the object
float CRATE_PHYSICS_DAMPING =				50.0f;		// amount of damping

float CRATE_SCALE =							2.0f;		// display scale
float CRATE_COLLIDE_PICKUP_SIZE =			150.0f;		// pickup size of the object
float CRATE_COLLIDE_NEAR_SIZE =				300.0f;		// near (run haptics) size of the object
float CRATE_HAPTICS_SIZE_X =				18.0f;		// haptics box size
float CRATE_HAPTICS_SIZE_Y =				8.0f;
float CRATE_HAPTICS_SIZE_Z =				8.0f;

// Car Physics constants
float CAR_FRONT_AXIS_DIST =					30.0f;		// distance from car center to front axle
float CAR_BACK_AXIS_DIST =					-60.0f;		// distance from car center to back axle
float CAR_WHEEL_LAT_DIST_L =				36.0f;		// distance out from car center to wheel (left wheel)
float CAR_WHEEL_LAT_DIST_R =				-40.0f;		// distance out from car center to wheel (right wheel)
float CAR_WHEEL_HEIGHT =					-14.0f;		// height of wheels relative to car center

float CAR_PHYSICS_MASS_WHEEL =				30.0f;		// mass per wheel
float CAR_PHYSICS_MASS_CHASSIS =			5.0f;		// mass for chassis lighter than wheels so it doesn't tilt

float CAR_PHYSICS_FRICTION_WHEEL =			100.0f;		// friction of wheels
float CAR_PHYSICS_FRICTION_CHASSIS =		5.0f;		// friction of chassis (low to avoid getting stuck on ramps)

float CAR_PHYSICS_DAMPING_WHEEL =			40.0f;		// damping ang for wheels
float CAR_PHYSICS_DAMPING_CHASSIS =			100.0f;		// damping lin/ang for chassis

float CAR_PHYSICS_SUSPENSION_ERP =			95000.0f;	// how quickly shocks are reduced to correct position (100,000 max, 0: never)
float CAR_PHYSICS_SUSPENSION_CFM =			250.0f;		// how far shocks can overshoot (0: never, 100,000 max)

float CAR_PHYSICS_MAX_TORQUE =				800.0f;		// maximum torque applied to wheels
float CAR_PHYSICS_MAX_TORQUE_BRAKE =		1000.0f;	// maximum torque when braking

float CAR_PHYSICS_STEER_SCALE =				0.5f;		// steering scale factor
float CAR_PHYSICS_MAX_STEER =				45.0f;		// max steer angle
float CAR_PHYSICS_DEADZONE_STEER =			30.0f;		// dead-zone at center of haptics control

float CAR_PHYSICS_THROTTLE_SCALE =			100.0f;		// throttle scale factor
float CAR_PHYSICS_MAX_THROTTLE =			1500.0f;	// max throttle
float CAR_PHYSICS_DEADZONE_THROTTLE =		10.0f;		// dead-zone at center of haptics control, throttle

float CAR_UPRIGHT_SPRING_SCALE =			5000.0f;	// scale factor for the spring that keeps the car upright

float CAR_HOLDING_VELOCITY_SCALE =			100.0f;		// velocity scale to use when holding the car

float CAR_COLLIDE_PICKUP_SIZE =				100.0f;		// pickup collision size
float CAR_COLLIDE_NEAR_SIZE =				200.0f;		// near to cursor haptics threshold
float CAR_HAPTICS_SIZE_X =					20.0f;		// haptics box size
float CAR_HAPTICS_SIZE_Y =					10.0f;
float CAR_HAPTICS_SIZE_Z =					8.0f;

// button values
float SHRINK_BTN_SANDBOX_X =				1000.0f;	// shrink button screen position
float SHRINK_BTN_SANDBOX_Y =				300.0f;
float SHRINK_BTN_SANDBOX_Z =				200.0f;

float SHRINK_BTN_SCALE =					3.0f;		// shrink button display scale
float SHRINKBTN_COLLIDE_SIZE =				100.0f;		// click distance
float SHRINKBTN_NEAR_SIZE =					200.0f;		// near (run haptics) size of the object
float SHRINKBTN_HAPTICS_SIZE =				17.0f;		// haptics radius

float GROW_BTN_SANDBOX_X =					40.0f;		// grow button screen position
float GROW_BTN_SANDBOX_Y =					10.0f;
float GROW_BTN_SANDBOX_Z =					-7.0f;

float GROW_BTN_SCALE =						0.15f;		// grow button display scale
float GROWBTN_COLLIDE_SIZE =				3.0f;		// click distance
float GROWBTN_NEAR_SIZE =					6.0f;		// near (run haptics) size of the object
float GROWBTN_HAPTICS_SIZE =				12.0f;		// haptics radius

float BTN_MIN_SCALE =						0.75f;		// minimum scale of the button (adjusted from 0..1.0)
float BTN_CLICK_SCALE =						0.85f;		// click scale of the button (adjusted from 0..1.0)


//---- haptics vars that can be tuned for each type of device----

// general values
static int    IS_FALCON = 0;
static double TIME_SCALE = 1;

// haptics object stiffnesses
double SURFACE_STIFFNESS =					0.95f;		// floor surfaces
double SPHERE_STIFFNESS =					0.6f;		// sphere haptics object
double BOX_STIFFNESS =						0.6f;		// box haptics object
double DRIVING_STIFFNESS =					0.02f;		// driving constraint
double BUTTON_STIFFNESS =					0.5f;		// button
double BUTTON_CLICK_STIFFNESS =				0.2f;		// button when clicked

// sandbox view values
float TOOL_SANDBOX_OFFSET_POS_X =			0.0f;		// Offset the X position of the hand tool, sandbox view
float TOOL_SANDBOX_OFFSET_POS_Y =			0.0f;		// Offset the Y position of the hand tool, sandbox view
float TOOL_SANDBOX_OFFSET_POS_Z =			1500.0f;	// Offset the Z position of the hand tool, sandbox view
float TOOL_SANDBOX_MOVEMENT_SCALE_BASE =	4.0f;		// Scale the base movement of the hand tool, sandbox view
float TOOL_SANDBOX_MOVEMENT_SCALE_Z =		2.5f;		// additional Z Scale of the hand tool, sandbox view

// driving view values
float TOOL_DRIVING_OFFSET_POS_X =			0.0f;		// Offset the X position of the hand tool, driving view
float TOOL_DRIVING_OFFSET_POS_Y =			0.0f;		// Offset the Y position of the hand tool, driving view
float TOOL_DRIVING_OFFSET_POS_Z =			70.0f;		// Offset the Z position of the hand tool, driving view
float TOOL_DRIVING_MOVEMENT_SCALE_BASE =	0.2f;		// Scale the base movement of the hand tool, driving view
float TOOL_DRIVING_MOVEMENT_SCALE_Z =		2.5f;		// additional Z Scale of the hand tool, driving view

float DRIVING_CONSTRAINT_OFFSET_X =			0.0f;		// driving mode constrained forces offset
float DRIVING_CONSTRAINT_OFFSET_Y =			0.0f;
float DRIVING_CONSTRAINT_OFFSET_Z =			0.0f;

float DRIVING_CONSTRAINT_SCALE_X =			0.000003f;	// driving mode constrained forces scale
float DRIVING_CONSTRAINT_SCALE_Y =			0.001f;
float DRIVING_CONSTRAINT_SCALE_Z =			0.0001f;

float CHASSIS_IMPACT_THRESHOLD =			10.0f;		// speed threshold that causes the chassis to generate a haptics impact force
float CHASSIS_IMPACT_SCALE =				1.0f;		// scale to apply to chassis impact haptics forces

float WHEEL_IMPACT_THRESHOLD =				70.0f;		// speed threshold that causes a wheel to generate a haptics impact force
float WHEEL_IMPACT_SCALE =					0.2f;		// scale to apply to wheel impact haptics forces

// object weights
float FLOWERPOT_HAPTICS_WEIGHT =			2.0f;		// haptics weight
float CRATE_HAPTICS_WEIGHT =				2.0f;		// haptics weight
float CAR_HAPTICS_WEIGHT =					1.5f;		// the haptics pickup weight

//-------------------------------------------------


namespace 
{
double getSystemTime()
{
    static double s_wavelength = 0.0;
    static double s_wavelength_x_high = 0.0;
    static bool s_isFirstTime = true;

    if (s_isFirstTime)
    {
        LARGE_INTEGER l_frequency = { 0 };
        ::QueryPerformanceFrequency(&l_frequency);
        s_wavelength = 1.0 / double(l_frequency.LowPart);
        s_wavelength_x_high = s_wavelength * double(MAXDWORD);
        s_isFirstTime = false;
    }

    LARGE_INTEGER l_ticks = { 0 };
    ::QueryPerformanceCounter(&l_ticks);

    return s_wavelength_x_high * double(l_ticks.HighPart) +
        s_wavelength * double(l_ticks.LowPart);
}
};


// Main callback that runs the dynamics simulation within the haptics loop.
// HD API specific code.
HDLServoOpExitCode ContactCB(void *data)
{
    haptics* l_pHaptics = static_cast< haptics* >( data );
    if ( !l_pHaptics->m_isServoRunning ) return HDL_SERVOOP_CONTINUE;
 
    double position[3];
    bool button;

    hdlToolPosition(position);
    hdlToolButton(&button);
    
    l_pHaptics->m_terminusPositionServo = 
        etVector(position[0] * SCALE_FACTOR, 
                 position[1] * SCALE_FACTOR, 
                 position[2] * SCALE_FACTOR);
    l_pHaptics->m_isButtonDownServo = button;

    // query force from forceField of tObject        
    etVector l_forceVec = l_pHaptics->calculateForces();
    double l_force[3];
    l_force[0] = l_forceVec.x();
    l_force[1] = l_forceVec.y();
    l_force[2] = l_forceVec.z();        
    
    // send force to device
    hdlSetToolForce( l_force );

    return HDL_SERVOOP_CONTINUE;
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
void haptics::create()
{
    assert(!s_instance);
    s_instance = new haptics();
}


void haptics::destroy()
{
    assert(s_instance);
    delete s_instance;
}


haptics* haptics::instance()
{
    return s_instance;
}


void haptics::start() 
{ 
    m_isServoRunning = true; 
}


// **********
// Constructor
// **********
haptics::haptics()
    : m_hContactCB(HDL_INVALID_HANDLE),
      m_terminusPositionServo(0,0,0),
      m_terminusPositionClient(0,0,0),
      m_isButtonDownServo(false),
      m_isButtonDownClient(false),    
      m_isFirstTime(true),
      m_lastTime(0),
      m_prevToolPosition(0,0,0),
	  m_numSurfaces(0),
	  m_numSpheres(0),
	  m_numBoxes(0),
	  m_numForces(0),
	  m_bDoDrivingForces(0),
	  m_log()	
	
{
#ifdef LOG
    fd = fopen("tr1haptic.txt", "w");
#endif

    // Initialize basic scene graph (scene and root)
    initialize();
}


// **********
// Destructor
// **********
haptics::~haptics()
{
#ifdef LOG
    fclose(fd);
#endif
    hdlDestroyServoOp( m_hContactCB );
    hdlUninitDevice( g_hDevice );
    g_hDevice = HDL_INVALID_HANDLE;

	s_instance = 0;
}


// **********
// Does first order initialization of GHOST scene graph.
// Scene and Root nodes are set up.
// **********
bool haptics::initialize()
{
    // Initialize device
    const HDLDeviceID deviceIDFiller = 0;
    g_hDevice = hdlInitDevice(deviceIDFiller);

    hdlMakeCurrent(g_hDevice);
    hdlDeviceWorkspace(m_deviceWorkspaceDim);

    m_hContactCB = hdlCreateServoOp(ContactCB, this, bNonBlocking);

	m_holdingWeight = -1.0;

    initParams();

    return (true);
}

// Callback to sync data from haptics thread to client (graphics) thread.
HDLServoOpExitCode GetStateCB(void *pUsrData)
{
    haptics* l_pHaptics = static_cast<haptics*>(pUsrData);
    l_pHaptics->doSyncOp();

    return HDL_SERVOOP_EXIT;
}


void haptics::doSyncOp()
{
    m_terminusPositionClient = m_terminusPositionServo;
    m_terminusTransformClient = m_terminusTransformServo;
    m_isButtonDownClient = m_isButtonDownServo;
}


// Sets the state returned by haptics class
// in a thread safe manner by scheduling
// a copy in the servo thread.
void haptics::syncFromServo()
{
    hdlCreateServoOp(GetStateCB, this, bBlocking);
}


// **********
// This function returns the device position in world coordinates
// in millimeters.  World coordinates are those of the PHANTOM.
// Note that 0,0,0 is at the middle of the haptic devices
// workspace (i.e., the reset position)
// **********
bool haptics::getDeviceTerminusPositionWC(double &x, double &y, double &z)
{

    // Get the position of the haptic device
    // Note that as a default the coordinates are in millimeters
    // and 0,0,0 is a the middle of the haptic devices workspace
    // (i.e., its reset position).

	etVector posWC = m_terminusPositionClient;
	 
    x = posWC.x();
    y = posWC.y();
    z = posWC.z();
        
    return (true);
}


// Returns the orientation of the haptic device terminus in degrees
// If device has not been initialized, 0,0,0 is returned
bool haptics::getDeviceTerminusMatrixWC(etMatrix& cummatrix)

{
    cummatrix = etMatrix(m_terminusTransformClient);

    return true;
}


// Returns the state of the device (stylus) switch -- true if depressed, false otherwise
bool haptics::isDeviceSwitchDown()
{
    return m_isButtonDownClient;
}


double haptics::getParam(std::string a_paramName, double a_default)
{
    const int bufferSize = 100;
    char buffer[bufferSize];
    char l_default[100];
    sprintf(l_default, "%f", a_default);
    char pathBuf[MAX_PATH];
    _getcwd(pathBuf, MAX_PATH);
    strcat(pathBuf, "\\");
    strncat(pathBuf, iniFileName, strlen(iniFileName));
    GetPrivateProfileString(g_deviceType, a_paramName.c_str(), l_default,
                            buffer, bufferSize, pathBuf);
    return atof(buffer);
}

int haptics::getParam(std::string a_paramName, int a_default)
{
    const int bufferSize = 100;
    char buffer[bufferSize];
    char l_default[100];
    sprintf(l_default, "%d", a_default);
    char pathBuf[MAX_PATH];
    _getcwd(pathBuf, MAX_PATH);
    strcat(pathBuf, "\\");
    strncat(pathBuf, iniFileName, strlen(iniFileName));
    GetPrivateProfileString(g_deviceType, a_paramName.c_str(), l_default,
                            buffer, bufferSize, pathBuf);
    return atoi(buffer);
}

void haptics::initParams()
{
	// setup INI type to grab general vars from the haptics.ini file
	g_deviceType = "GENERAL";

	// grab all the general tuning vars (overriding default settings from the haptics.ini file settings)

	SANDBOX_CAMERA_DIST_BEHIND = getParam( "SANDBOX_CAMERA_DIST_BEHIND", SANDBOX_CAMERA_DIST_BEHIND );
	SANDBOX_CAMERA_DIST_ABOVE = getParam( "SANDBOX_CAMERA_DIST_ABOVE", SANDBOX_CAMERA_DIST_ABOVE );
	CAR_CAMERA_1ST_DIST_BEHIND = getParam( "CAR_CAMERA_1ST_DIST_BEHIND", CAR_CAMERA_1ST_DIST_BEHIND );
	CAR_CAMERA_1ST_DIST_ABOVE = getParam( "CAR_CAMERA_1ST_DIST_ABOVE", CAR_CAMERA_1ST_DIST_ABOVE );
	CAR_CAMERA_3RD_DIST_BEHIND = getParam( "CAR_CAMERA_3RD_DIST_BEHIND", CAR_CAMERA_3RD_DIST_BEHIND );
	CAR_CAMERA_3RD_DIST_ABOVE = getParam( "CAR_CAMERA_3RD_DIST_ABOVE", CAR_CAMERA_3RD_DIST_ABOVE );

	CAMERA_MOVE_TIME = getParam( "CAMERA_MOVE_TIME", CAMERA_MOVE_TIME);

	PHYSICS_GRAVITY = getParam( "PHYSICS_GRAVITY", PHYSICS_GRAVITY);

	OBJ_HOLDING_VELOCITY_SCALE = getParam( "OBJ_HOLDING_VELOCITY_SCALE", OBJ_HOLDING_VELOCITY_SCALE );

	FLOWERPOT_PHYSICS_MASS = getParam( "FLOWERPOT_PHYSICS_MASS", FLOWERPOT_PHYSICS_MASS );
	FLOWERPOT_PHYSICS_FRICTION = getParam( "FLOWERPOT_PHYSICS_FRICTION", FLOWERPOT_PHYSICS_FRICTION );
	FLOWERPOT_PHYSICS_DAMPING = getParam( "FLOWERPOT_PHYSICS_DAMPING", FLOWERPOT_PHYSICS_DAMPING );
	FLOWERPOT_COLLIDE_PICKUP_SIZE = getParam( "FLOWERPOT_COLLIDE_PICKUP_SIZE", FLOWERPOT_COLLIDE_PICKUP_SIZE );
	FLOWERPOT_COLLIDE_NEAR_SIZE = getParam( "FLOWERPOT_COLLIDE_NEAR_SIZE", FLOWERPOT_COLLIDE_NEAR_SIZE );
	FLOWERPOT_HAPTICS_RADIUS = getParam( "FLOWERPOT_HAPTICS_RADIUS", FLOWERPOT_HAPTICS_RADIUS );

	CRATE_PHYSICS_MASS = getParam( "CRATE_PHYSICS_MASS", CRATE_PHYSICS_MASS );
	CRATE_PHYSICS_FRICTION = getParam( "CRATE_PHYSICS_FRICTION", CRATE_PHYSICS_FRICTION );
	CRATE_PHYSICS_DAMPING = getParam( "CRATE_PHYSICS_DAMPING", CRATE_PHYSICS_DAMPING );
	CRATE_COLLIDE_PICKUP_SIZE = getParam( "CRATE_COLLIDE_PICKUP_SIZE", CRATE_COLLIDE_PICKUP_SIZE );
	CRATE_COLLIDE_NEAR_SIZE = getParam( "CRATE_COLLIDE_NEAR_SIZE", CRATE_COLLIDE_NEAR_SIZE );
	CRATE_HAPTICS_SIZE_X = getParam( "CRATE_HAPTICS_SIZE_X", CRATE_HAPTICS_SIZE_X );
	CRATE_HAPTICS_SIZE_Y = getParam( "CRATE_HAPTICS_SIZE_Y", CRATE_HAPTICS_SIZE_Y );
	CRATE_HAPTICS_SIZE_Z = getParam( "CRATE_HAPTICS_SIZE_Z", CRATE_HAPTICS_SIZE_Z );

	CAR_FRONT_AXIS_DIST = getParam( "CAR_FRONT_AXIS_DIST", CAR_FRONT_AXIS_DIST );
	CAR_BACK_AXIS_DIST = getParam( "CAR_BACK_AXIS_DIST", CAR_BACK_AXIS_DIST );
	CAR_WHEEL_LAT_DIST_L = getParam( "CAR_WHEEL_LAT_DIST_L", CAR_WHEEL_LAT_DIST_L );
	CAR_WHEEL_LAT_DIST_R = getParam( "CAR_WHEEL_LAT_DIST_R", CAR_WHEEL_LAT_DIST_R );
	CAR_WHEEL_HEIGHT = getParam( "CAR_WHEEL_HEIGHT", CAR_WHEEL_HEIGHT );
	CAR_PHYSICS_MASS_WHEEL = getParam( "CAR_PHYSICS_MASS_WHEEL", CAR_PHYSICS_MASS_WHEEL );
	CAR_PHYSICS_MASS_CHASSIS = getParam( "CAR_PHYSICS_MASS_CHASSIS", CAR_PHYSICS_MASS_CHASSIS );
	CAR_PHYSICS_FRICTION_WHEEL = getParam( "CAR_PHYSICS_FRICTION_WHEEL", CAR_PHYSICS_FRICTION_WHEEL );
	CAR_PHYSICS_FRICTION_CHASSIS = getParam( "CAR_PHYSICS_FRICTION_CHASSIS", CAR_PHYSICS_FRICTION_CHASSIS );
	CAR_PHYSICS_DAMPING_WHEEL = getParam( "CAR_PHYSICS_DAMPING_WHEEL", CAR_PHYSICS_DAMPING_WHEEL );
	CAR_PHYSICS_DAMPING_CHASSIS = getParam( "CAR_PHYSICS_DAMPING_CHASSIS", CAR_PHYSICS_DAMPING_CHASSIS );
	CAR_PHYSICS_SUSPENSION_ERP = getParam( "CAR_PHYSICS_SUSPENSION_ERP", CAR_PHYSICS_SUSPENSION_ERP );
	CAR_PHYSICS_SUSPENSION_CFM = getParam( "CAR_PHYSICS_SUSPENSION_CFM", CAR_PHYSICS_SUSPENSION_CFM );
	CAR_PHYSICS_MAX_TORQUE = getParam( "CAR_PHYSICS_MAX_TORQUE", CAR_PHYSICS_MAX_TORQUE );
	CAR_PHYSICS_MAX_TORQUE_BRAKE = getParam( "CAR_PHYSICS_MAX_TORQUE_BRAKE", CAR_PHYSICS_MAX_TORQUE_BRAKE );
	CAR_PHYSICS_STEER_SCALE = getParam( "CAR_PHYSICS_STEER_SCALE", CAR_PHYSICS_STEER_SCALE );
	CAR_PHYSICS_MAX_STEER = getParam( "CAR_PHYSICS_MAX_STEER", CAR_PHYSICS_MAX_STEER );
	CAR_PHYSICS_DEADZONE_STEER = getParam( "CAR_PHYSICS_DEADZONE_STEER", CAR_PHYSICS_DEADZONE_STEER );
	CAR_PHYSICS_THROTTLE_SCALE = getParam( "CAR_PHYSICS_THROTTLE_SCALE", CAR_PHYSICS_THROTTLE_SCALE );
	CAR_PHYSICS_MAX_THROTTLE = getParam( "CAR_PHYSICS_MAX_THROTTLE", CAR_PHYSICS_MAX_THROTTLE );
	CAR_PHYSICS_DEADZONE_THROTTLE = getParam( "CAR_PHYSICS_DEADZONE_THROTTLE", CAR_PHYSICS_DEADZONE_THROTTLE );
	CAR_UPRIGHT_SPRING_SCALE = getParam( "CAR_UPRIGHT_SPRING_SCALE", CAR_UPRIGHT_SPRING_SCALE );
	CAR_HOLDING_VELOCITY_SCALE = getParam( "CAR_HOLDING_VELOCITY_SCALE", CAR_HOLDING_VELOCITY_SCALE );
	CAR_COLLIDE_PICKUP_SIZE = getParam( "CAR_COLLIDE_PICKUP_SIZE", CAR_COLLIDE_PICKUP_SIZE );
	CAR_COLLIDE_NEAR_SIZE = getParam( "CAR_COLLIDE_NEAR_SIZE", CAR_COLLIDE_NEAR_SIZE );
	CAR_HAPTICS_SIZE_X = getParam( "CAR_HAPTICS_SIZE_X", CAR_HAPTICS_SIZE_X );
	CAR_HAPTICS_SIZE_Y = getParam( "CAR_HAPTICS_SIZE_Y", CAR_HAPTICS_SIZE_Y );
	CAR_HAPTICS_SIZE_Z = getParam( "CAR_HAPTICS_SIZE_Z", CAR_HAPTICS_SIZE_Z );

	SHRINK_BTN_SANDBOX_X = getParam( "SHRINK_BTN_SANDBOX_X", SHRINK_BTN_SANDBOX_X );
	SHRINK_BTN_SANDBOX_Y = getParam( "SHRINK_BTN_SANDBOX_Y", SHRINK_BTN_SANDBOX_Y );
	SHRINK_BTN_SANDBOX_Z = getParam( "SHRINK_BTN_SANDBOX_Z", SHRINK_BTN_SANDBOX_Z );
	SHRINK_BTN_SCALE = getParam( "SHRINK_BTN_SCALE", SHRINK_BTN_SCALE );
	SHRINKBTN_COLLIDE_SIZE = getParam( "SHRINKBTN_COLLIDE_SIZE", SHRINKBTN_COLLIDE_SIZE );
	SHRINKBTN_NEAR_SIZE = getParam( "SHRINKBTN_NEAR_SIZE", SHRINKBTN_NEAR_SIZE );
	SHRINKBTN_HAPTICS_SIZE = getParam( "SHRINKBTN_HAPTICS_SIZE", SHRINKBTN_HAPTICS_SIZE );
		
	GROW_BTN_SANDBOX_X = getParam( "GROW_BTN_SANDBOX_X", GROW_BTN_SANDBOX_X );
	GROW_BTN_SANDBOX_Y = getParam( "GROW_BTN_SANDBOX_Y", GROW_BTN_SANDBOX_Y );
	GROW_BTN_SANDBOX_Z = getParam( "GROW_BTN_SANDBOX_Z", GROW_BTN_SANDBOX_Z );
	GROW_BTN_SCALE = getParam( "GROW_BTN_SCALE", GROW_BTN_SCALE );
	GROWBTN_COLLIDE_SIZE = getParam( "GROWBTN_COLLIDE_SIZE", GROWBTN_COLLIDE_SIZE );
	GROWBTN_NEAR_SIZE = getParam( "GROWBTN_NEAR_SIZE", GROWBTN_NEAR_SIZE );
	GROWBTN_HAPTICS_SIZE = getParam( "GROWBTN_HAPTICS_SIZE", GROWBTN_HAPTICS_SIZE );

	BTN_MIN_SCALE = getParam( "BTN_MIN_SCALE", BTN_MIN_SCALE );
	BTN_CLICK_SCALE = getParam( "BTN_CLICK_SCALE", BTN_CLICK_SCALE );

	// now set to grab .ini vars for the specific haptics type
	// NOTE: should usually be FALCON_ALPHA (for a Novint Falcon)
	//		 I've left tuning values for the PHANTOM_OMNI in the INI file, however, since that was the device I originally tested this with
    g_deviceType = hdlDeviceModel();

    // Override default settings if specified in a haptics.ini file
    IS_FALCON = getParam( "IS_FALCON", IS_FALCON );
    TIME_SCALE = getParam( "TIME_SCALE", TIME_SCALE );

	SURFACE_STIFFNESS = getParam( "SURFACE_STIFFNESS", SURFACE_STIFFNESS );
    SPHERE_STIFFNESS = getParam( "SPHERE_STIFFNESS", SPHERE_STIFFNESS );
    BOX_STIFFNESS = getParam( "BOX_STIFFNESS", BOX_STIFFNESS );
    DRIVING_STIFFNESS = getParam( "DRIVING_STIFFNESS", DRIVING_STIFFNESS );
	BUTTON_STIFFNESS = getParam( "BUTTON_STIFFNESS", BUTTON_STIFFNESS );
	BUTTON_CLICK_STIFFNESS = getParam( "BUTTON_CLICK_STIFFNESS", BUTTON_CLICK_STIFFNESS );

	TOOL_SANDBOX_OFFSET_POS_X = getParam( "TOOL_SANDBOX_OFFSET_POS_X", TOOL_SANDBOX_OFFSET_POS_X );
    TOOL_SANDBOX_OFFSET_POS_Y = getParam( "TOOL_SANDBOX_OFFSET_POS_Y", TOOL_SANDBOX_OFFSET_POS_Y );
    TOOL_SANDBOX_OFFSET_POS_Z = getParam( "TOOL_SANDBOX_OFFSET_POS_Z", TOOL_SANDBOX_OFFSET_POS_Z );
    TOOL_SANDBOX_MOVEMENT_SCALE_BASE = getParam( "TOOL_SANDBOX_MOVEMENT_SCALE_BASE", TOOL_SANDBOX_MOVEMENT_SCALE_BASE );
    TOOL_SANDBOX_MOVEMENT_SCALE_Z = getParam( "TOOL_SANDBOX_MOVEMENT_SCALE_Z", TOOL_SANDBOX_MOVEMENT_SCALE_Z );

	TOOL_DRIVING_OFFSET_POS_X = getParam( "TOOL_DRIVING_OFFSET_POS_X", TOOL_DRIVING_OFFSET_POS_X );
    TOOL_DRIVING_OFFSET_POS_Y = getParam( "TOOL_DRIVING_OFFSET_POS_Y", TOOL_DRIVING_OFFSET_POS_Y );
    TOOL_DRIVING_OFFSET_POS_Z = getParam( "TOOL_DRIVING_OFFSET_POS_Z", TOOL_DRIVING_OFFSET_POS_Z );
    TOOL_DRIVING_MOVEMENT_SCALE_BASE = getParam( "TOOL_DRIVING_MOVEMENT_SCALE_BASE", TOOL_DRIVING_MOVEMENT_SCALE_BASE );
    TOOL_DRIVING_MOVEMENT_SCALE_Z = getParam( "TOOL_DRIVING_MOVEMENT_SCALE_Z", TOOL_DRIVING_MOVEMENT_SCALE_Z );

	DRIVING_CONSTRAINT_OFFSET_X = getParam( "DRIVING_CONSTRAINT_OFFSET_X", DRIVING_CONSTRAINT_OFFSET_X );
	DRIVING_CONSTRAINT_OFFSET_Y = getParam( "DRIVING_CONSTRAINT_OFFSET_Y", DRIVING_CONSTRAINT_OFFSET_Y );
	DRIVING_CONSTRAINT_OFFSET_Z = getParam( "DRIVING_CONSTRAINT_OFFSET_Z", DRIVING_CONSTRAINT_OFFSET_Z );
	DRIVING_CONSTRAINT_SCALE_X = getParam( "DRIVING_CONSTRAINT_SCALE_X", DRIVING_CONSTRAINT_SCALE_X );
	DRIVING_CONSTRAINT_SCALE_Y = getParam( "DRIVING_CONSTRAINT_SCALE_Y", DRIVING_CONSTRAINT_SCALE_Y );
	DRIVING_CONSTRAINT_SCALE_Z = getParam( "DRIVING_CONSTRAINT_SCALE_Z", DRIVING_CONSTRAINT_SCALE_Z );

	CHASSIS_IMPACT_THRESHOLD = getParam( "CHASSIS_IMPACT_THRESHOLD", CHASSIS_IMPACT_THRESHOLD );
	CHASSIS_IMPACT_SCALE = getParam( "CHASSIS_IMPACT_SCALE", CHASSIS_IMPACT_SCALE );
	WHEEL_IMPACT_THRESHOLD = getParam( "WHEEL_IMPACT_THRESHOLD", WHEEL_IMPACT_THRESHOLD );
	WHEEL_IMPACT_SCALE = getParam( "WHEEL_IMPACT_SCALE", WHEEL_IMPACT_SCALE );

	FLOWERPOT_HAPTICS_WEIGHT = getParam( "FLOWERPOT_HAPTICS_WEIGHT", FLOWERPOT_HAPTICS_WEIGHT );
	CRATE_HAPTICS_WEIGHT = getParam( "CRATE_HAPTICS_WEIGHT", CRATE_HAPTICS_WEIGHT );
	CAR_HAPTICS_WEIGHT = getParam( "CAR_HAPTICS_WEIGHT", CAR_HAPTICS_WEIGHT );

#ifdef LOG
	// print out the final settings for all tuning values to the LOG file....

	fprintf(fd, "----GENERAL TUNING VARS----\n");
	fprintf(fd, "SANDBOX_CAMERA_DIST_BEHIND = %f\n", SANDBOX_CAMERA_DIST_BEHIND);
	fprintf(fd, "SANDBOX_CAMERA_DIST_ABOVE = %f\n", SANDBOX_CAMERA_DIST_ABOVE);
	fprintf(fd, "CAR_CAMERA_1ST_DIST_BEHIND = %f\n", CAR_CAMERA_1ST_DIST_BEHIND);
	fprintf(fd, "CAR_CAMERA_1ST_DIST_ABOVE = %f\n", CAR_CAMERA_1ST_DIST_ABOVE);
	fprintf(fd, "CAR_CAMERA_3RD_DIST_BEHIND = %f\n", CAR_CAMERA_3RD_DIST_BEHIND);
	fprintf(fd, "CAR_CAMERA_3RD_DIST_ABOVE = %f\n", CAR_CAMERA_3RD_DIST_ABOVE);

	fprintf(fd, "CAMERA_MOVE_TIME = %f\n", CAMERA_MOVE_TIME);

	fprintf(fd, "PHYSICS_GRAVITY = %f\n", PHYSICS_GRAVITY);

	fprintf(fd, "OBJ_HOLDING_VELOCITY_SCALE = %f\n", OBJ_HOLDING_VELOCITY_SCALE);

	fprintf(fd, "FLOWERPOT_PHYSICS_MASS = %f\n", FLOWERPOT_PHYSICS_MASS);
	fprintf(fd, "FLOWERPOT_PHYSICS_FRICTION = %f\n", FLOWERPOT_PHYSICS_FRICTION);
	fprintf(fd, "FLOWERPOT_PHYSICS_DAMPING = %f\n", FLOWERPOT_PHYSICS_DAMPING);
	fprintf(fd, "FLOWERPOT_COLLIDE_PICKUP_SIZE = %f\n", FLOWERPOT_COLLIDE_PICKUP_SIZE);
	fprintf(fd, "FLOWERPOT_COLLIDE_NEAR_SIZE = %f\n", FLOWERPOT_COLLIDE_NEAR_SIZE);
	fprintf(fd, "FLOWERPOT_HAPTICS_RADIUS = %f\n", FLOWERPOT_HAPTICS_RADIUS);

	fprintf(fd, "CRATE_PHYSICS_MASS = %f\n", CRATE_PHYSICS_MASS);
	fprintf(fd, "CRATE_PHYSICS_FRICTION = %f\n", CRATE_PHYSICS_FRICTION);
	fprintf(fd, "CRATE_PHYSICS_DAMPING = %f\n", CRATE_PHYSICS_DAMPING);
	fprintf(fd, "CRATE_COLLIDE_PICKUP_SIZE = %f\n", CRATE_COLLIDE_PICKUP_SIZE);
	fprintf(fd, "CRATE_COLLIDE_NEAR_SIZE = %f\n", CRATE_COLLIDE_NEAR_SIZE);
	fprintf(fd, "CRATE_HAPTICS_SIZE_X = %f\n", CRATE_HAPTICS_SIZE_X);
	fprintf(fd, "CRATE_HAPTICS_SIZE_Y = %f\n", CRATE_HAPTICS_SIZE_Y);
	fprintf(fd, "CRATE_HAPTICS_SIZE_Z = %f\n", CRATE_HAPTICS_SIZE_Z);

	fprintf(fd, "CAR_FRONT_AXIS_DIST = %f\n", CAR_FRONT_AXIS_DIST);
	fprintf(fd, "CAR_BACK_AXIS_DIST = %f\n", CAR_BACK_AXIS_DIST);
	fprintf(fd, "CAR_WHEEL_LAT_DIST_L = %f\n", CAR_WHEEL_LAT_DIST_L);
	fprintf(fd, "CAR_WHEEL_LAT_DIST_R = %f\n", CAR_WHEEL_LAT_DIST_R);
	fprintf(fd, "CAR_WHEEL_HEIGHT = %f\n", CAR_WHEEL_HEIGHT);
	fprintf(fd, "CAR_PHYSICS_MASS_WHEEL = %f\n", CAR_PHYSICS_MASS_WHEEL);
	fprintf(fd, "CAR_PHYSICS_MASS_CHASSIS = %f\n", CAR_PHYSICS_MASS_CHASSIS);
	fprintf(fd, "CAR_PHYSICS_FRICTION_WHEEL = %f\n", CAR_PHYSICS_FRICTION_WHEEL);
	fprintf(fd, "CAR_PHYSICS_FRICTION_CHASSIS = %f\n", CAR_PHYSICS_FRICTION_CHASSIS);
	fprintf(fd, "CAR_PHYSICS_DAMPING_WHEEL = %f\n", CAR_PHYSICS_DAMPING_WHEEL);
	fprintf(fd, "CAR_PHYSICS_DAMPING_CHASSIS = %f\n", CAR_PHYSICS_DAMPING_CHASSIS);
	fprintf(fd, "CAR_PHYSICS_SUSPENSION_ERP = %f\n", CAR_PHYSICS_SUSPENSION_ERP);
	fprintf(fd, "CAR_PHYSICS_SUSPENSION_CFM = %f\n", CAR_PHYSICS_SUSPENSION_CFM);
	fprintf(fd, "CAR_PHYSICS_MAX_TORQUE = %f\n", CAR_PHYSICS_MAX_TORQUE);
	fprintf(fd, "CAR_PHYSICS_MAX_TORQUE_BRAKE = %f\n", CAR_PHYSICS_MAX_TORQUE_BRAKE);
	fprintf(fd, "CAR_PHYSICS_STEER_SCALE = %f\n", CAR_PHYSICS_STEER_SCALE);
	fprintf(fd, "CAR_PHYSICS_MAX_STEER = %f\n", CAR_PHYSICS_MAX_STEER);
	fprintf(fd, "CAR_PHYSICS_DEADZONE_STEER = %f\n", CAR_PHYSICS_DEADZONE_STEER);
	fprintf(fd, "CAR_PHYSICS_THROTTLE_SCALE = %f\n", CAR_PHYSICS_THROTTLE_SCALE);
	fprintf(fd, "CAR_PHYSICS_MAX_THROTTLE = %f\n", CAR_PHYSICS_MAX_THROTTLE);
	fprintf(fd, "CAR_PHYSICS_DEADZONE_THROTTLE = %f\n", CAR_PHYSICS_DEADZONE_THROTTLE);
	fprintf(fd, "CAR_UPRIGHT_SPRING_SCALE = %f\n", CAR_UPRIGHT_SPRING_SCALE);
	fprintf(fd, "CAR_HOLDING_VELOCITY_SCALE = %f\n", CAR_HOLDING_VELOCITY_SCALE);
	fprintf(fd, "CAR_COLLIDE_PICKUP_SIZE = %f\n", CAR_COLLIDE_PICKUP_SIZE);
	fprintf(fd, "CAR_COLLIDE_NEAR_SIZE = %f\n", CAR_COLLIDE_NEAR_SIZE);
	fprintf(fd, "CAR_HAPTICS_SIZE_X = %f\n", CAR_HAPTICS_SIZE_X);
	fprintf(fd, "CAR_HAPTICS_SIZE_Y = %f\n", CAR_HAPTICS_SIZE_Y);
	fprintf(fd, "CAR_HAPTICS_SIZE_Z = %f\n", CAR_HAPTICS_SIZE_Z);

	fprintf(fd, "SHRINK_BTN_SANDBOX_X = %f\n", SHRINK_BTN_SANDBOX_X);
	fprintf(fd, "SHRINK_BTN_SANDBOX_Y = %f\n", SHRINK_BTN_SANDBOX_Y);
	fprintf(fd, "SHRINK_BTN_SANDBOX_Z = %f\n", SHRINK_BTN_SANDBOX_Z);
	fprintf(fd, "SHRINK_BTN_SCALE = %f\n", SHRINK_BTN_SCALE);
	fprintf(fd, "SHRINKBTN_COLLIDE_SIZE = %f\n", SHRINKBTN_COLLIDE_SIZE);
	fprintf(fd, "SHRINKBTN_NEAR_SIZE = %f\n", SHRINKBTN_NEAR_SIZE);
	fprintf(fd, "SHRINKBTN_HAPTICS_SIZE = %f\n", SHRINKBTN_HAPTICS_SIZE);
		
	fprintf(fd, "GROW_BTN_SANDBOX_X = %f\n", GROW_BTN_SANDBOX_X);
	fprintf(fd, "GROW_BTN_SANDBOX_Y = %f\n", GROW_BTN_SANDBOX_Y);
	fprintf(fd, "GROW_BTN_SANDBOX_Z = %f\n", GROW_BTN_SANDBOX_Z);
	fprintf(fd, "GROW_BTN_SCALE = %f\n", GROW_BTN_SCALE);
	fprintf(fd, "GROWBTN_COLLIDE_SIZE = %f\n", GROWBTN_COLLIDE_SIZE);
	fprintf(fd, "GROWBTN_NEAR_SIZE = %f\n", GROWBTN_NEAR_SIZE);
	fprintf(fd, "GROWBTN_HAPTICS_SIZE = %f\n", GROWBTN_HAPTICS_SIZE);

	fprintf(fd, "BTN_MIN_SCALE = %f\n", BTN_MIN_SCALE);
	fprintf(fd, "BTN_CLICK_SCALE = %f\n", BTN_CLICK_SCALE);

	fprintf(fd, "\n----HAPTICS DEVICE TYPE VARS----\n");
    fprintf(fd, "Device name = %s\n", hdlDeviceModel());

	fprintf(fd, "IS_FALCON = %d\n", IS_FALCON);
    fprintf(fd, "TIME_SCALE = %d\n", TIME_SCALE);

	fprintf(fd, "SURFACE_STIFFNESS = %f\n", SURFACE_STIFFNESS);
    fprintf(fd, "SPHERE_STIFFNESS = %f\n", SPHERE_STIFFNESS);
    fprintf(fd, "BOX_STIFFNESS = %f\n", BOX_STIFFNESS);
    fprintf(fd, "DRIVING_STIFFNESS = %f\n", DRIVING_STIFFNESS);
	fprintf(fd, "BUTTON_STIFFNESS = %f\n", BUTTON_STIFFNESS);
	fprintf(fd, "BUTTON_CLICK_STIFFNESS = %f\n", BUTTON_CLICK_STIFFNESS);

    fprintf(fd, "TOOL_SANDBOX_OFFSET_POS_X = %f\n", TOOL_SANDBOX_OFFSET_POS_X);
    fprintf(fd, "TOOL_SANDBOX_OFFSET_POS_Y = %f\n", TOOL_SANDBOX_OFFSET_POS_Y);
    fprintf(fd, "TOOL_SANDBOX_OFFSET_POS_Z = %f\n", TOOL_SANDBOX_OFFSET_POS_Z);
    fprintf(fd, "TOOL_SANDBOX_MOVEMENT_SCALE_BASE = %f\n", TOOL_SANDBOX_MOVEMENT_SCALE_BASE);
    fprintf(fd, "TOOL_SANDBOX_MOVEMENT_SCALE_Z = %f\n", TOOL_SANDBOX_MOVEMENT_SCALE_Z);

    fprintf(fd, "TOOL_DRIVING_OFFSET_POS_X = %f\n", TOOL_DRIVING_OFFSET_POS_X);
    fprintf(fd, "TOOL_DRIVING_OFFSET_POS_Y = %f\n", TOOL_DRIVING_OFFSET_POS_Y);
    fprintf(fd, "TOOL_DRIVING_OFFSET_POS_Z = %f\n", TOOL_DRIVING_OFFSET_POS_Z);
    fprintf(fd, "TOOL_DRIVING_MOVEMENT_SCALE_BASE = %f\n", TOOL_DRIVING_MOVEMENT_SCALE_BASE);
    fprintf(fd, "TOOL_DRIVING_MOVEMENT_SCALE_Z = %f\n", TOOL_DRIVING_MOVEMENT_SCALE_Z);

	fprintf(fd, "DRIVING_CONSTRAINT_OFFSET_X = %f\n", DRIVING_CONSTRAINT_OFFSET_X);
	fprintf(fd, "DRIVING_CONSTRAINT_OFFSET_Y = %f\n", DRIVING_CONSTRAINT_OFFSET_Y);
	fprintf(fd, "DRIVING_CONSTRAINT_OFFSET_Z = %f\n", DRIVING_CONSTRAINT_OFFSET_Z);
	fprintf(fd, "DRIVING_CONSTRAINT_SCALE_X = %f\n", DRIVING_CONSTRAINT_SCALE_X);
	fprintf(fd, "DRIVING_CONSTRAINT_SCALE_Y = %f\n", DRIVING_CONSTRAINT_SCALE_Y);
	fprintf(fd, "DRIVING_CONSTRAINT_SCALE_Z = %f\n", DRIVING_CONSTRAINT_SCALE_Z);

	fprintf(fd, "CHASSIS_IMPACT_THRESHOLD = %f", CHASSIS_IMPACT_THRESHOLD);
	fprintf(fd, "CHASSIS_IMPACT_SCALE = %f", CHASSIS_IMPACT_SCALE);
	fprintf(fd, "WHEEL_IMPACT_THRESHOLD = %f", WHEEL_IMPACT_THRESHOLD);
	fprintf(fd, "WHEEL_IMPACT_SCALE = %f", WHEEL_IMPACT_SCALE);

	fprintf(fd, "FLOWERPOT_HAPTICS_WEIGHT = %f\n", FLOWERPOT_HAPTICS_WEIGHT);
	fprintf(fd, "CRATE_HAPTICS_WEIGHT = %f\n", CRATE_HAPTICS_WEIGHT);
	fprintf(fd, "CAR_HAPTICS_WEIGHT = %f\n", CAR_HAPTICS_WEIGHT);
#endif
}


//********************************************************
// Create the haptic forces for a sphere
//
// IN:	a_position = the sphere position
//		a_radius = the radius of the sphere
//		a_toolPos = the position of the haptics device
//		a_stiffness = the stiffness of the sphere
//
// RETURNS: the force vector
//********************************************************
etVector haptics::doSphereAt(etVector a_position, double a_radius, etVector a_toolPos, double a_stiffness)
{
    etVector l_force(0, 0, 0);

    etVector l_posLC = a_toolPos - a_position;
    double l_dist = l_posLC.length();

	if (l_dist < a_radius)
    {
		double l_penetrationDist = a_radius - l_dist;
		double l_penetrationRatio = 1.0 - (l_penetrationDist / a_radius);

        l_force = l_posLC.normalize();

		l_force = l_force.scale(l_penetrationRatio * 2.0);
        l_force = l_force.scale(a_stiffness);
    }

    return l_force;
}


//********************************************************
// Create the haptic forces for an oriented box
//
// IN:	a_position = the box position
//		a_size = the size of the box
//		a_headvec = the heading vector of the box
//		a_upVec = the up vector of the box
//		a_toolPos = the position of the haptics device
//		a_stiffness = the stiffness of the box
//
// RETURNS: the force vector
//********************************************************
etVector haptics::doBoxAt(etVector a_position, etVector a_size, etVector a_headVec, etVector a_upVec, etVector a_toolPos, double a_stiffness)
{
	etVector l_force(0, 0, 0);

	etVector l_headVec;
	etVector l_wingVec;
	etVector l_upVec;

	etVector l_relPos;

	double l_relX;
	double l_relY;
	double l_relZ;

	double l_penX;
	double l_penY;
	double l_penZ;
	double l_penetration;

	int l_closestSide = 0;

	// create the 3 box axis
	l_headVec = a_headVec;

	l_wingVec = l_headVec.cross(a_upVec);

	l_upVec = l_headVec.cross(l_wingVec);

	l_headVec = l_headVec.normalize();
	l_wingVec = l_wingVec.normalize();
	l_upVec = l_upVec.normalize();

	// calculate the tool position relative to box pos/orientation
	l_relPos = a_toolPos - a_position;

	// dot prod each axis with the relative position, getting the 
	// projection along each axis
	l_relX = l_headVec.dot(l_relPos);
	l_relY = l_wingVec.dot(l_relPos);
	l_relZ = l_upVec.dot(l_relPos);

	if (   (l_relX > a_size.x()) || (l_relX < -a_size.x())
		|| (l_relY > a_size.y()) || (l_relY < -a_size.y())
		|| (l_relZ > a_size.z()) || (l_relZ < -a_size.z()))
	{	// outside the box... no forces
		return(l_force);
	}

	if (l_relX >= 0.0)
	{
		l_penX = a_size.x() - l_relX;
	}
	else
	{
		l_penX = -a_size.x() - l_relX;
	}

	if (l_relY >= 0.0)
	{
		l_penY = a_size.y() - l_relY;
	}
	else
	{
		l_penY = -a_size.y() - l_relY;
	}

	if (l_relZ >= 0.0)
	{
		l_penZ = a_size.z() - l_relZ;
	}
	else
	{
		l_penZ = -a_size.z() - l_relZ;
	}

	if (fabs(l_penX) < fabs(l_penY))
	{	// x is less than y
		if (fabs(l_penX) < fabs(l_penZ))
		{	// x is less than z also
			l_closestSide = 0;	// hit x axis side
			l_penetration = l_penX;
		}
		else if (fabs(l_penY) < fabs(l_penZ))
		{	// y is less than z, but greater than x
			l_closestSide = 0;	// hit x axis side
			l_penetration = l_penX;
		}
		else
		{	// z is smallest
			l_closestSide = 2;
			l_penetration = l_penZ;
		}
	}
	else
	{	// y is less than x
		if (fabs(l_penY) < fabs(l_penZ))
		{	// y is less than z also
			l_closestSide = 1;
			l_penetration = l_penY;
		}
		else if (fabs(l_penX) < fabs(l_penZ))
		{	// x is less than z, but greater than y
			l_closestSide = 1;
			l_penetration = l_penY;
		}
		else
		{	// z is smallest
			l_closestSide = 2;
			l_penetration = l_penZ;
		}
	}

	switch (l_closestSide)
	{
	case 0:
		l_force = l_headVec;
		break;

	case 1:
		l_force = l_wingVec;
		break;

	case 2:
		l_force = l_upVec;
		break;
	}
		
	l_force = l_force.scale(l_penetration);
	l_force = l_force.scale(a_stiffness);

	return(l_force);
}


//***********************************************************
// run the constrained driving forces
//
// IN: a_toolPos = the position of the haptics device
//
// RETURNS: the force vector
//***********************************************************
etVector haptics::doDrivingForces(etVector a_toolPos)
{
	etVector l_force(0, 0, 0);

	double l_adjustedPosX;
	double l_adjustedPosY;
	double l_adjustedPosZ;

	double l_forceX;
	double l_forceY;
	double l_forceZ;

	l_adjustedPosX = a_toolPos.x() + DRIVING_CONSTRAINT_OFFSET_X;
	l_adjustedPosY = a_toolPos.y() + DRIVING_CONSTRAINT_OFFSET_X;
	l_adjustedPosZ = a_toolPos.z() + DRIVING_CONSTRAINT_OFFSET_X;

	l_forceX = -l_adjustedPosX * l_adjustedPosX * l_adjustedPosX * DRIVING_CONSTRAINT_SCALE_X;
	l_forceY = -l_adjustedPosY * l_adjustedPosY * l_adjustedPosY * DRIVING_CONSTRAINT_SCALE_Y;
	l_forceZ = -l_adjustedPosZ * l_adjustedPosZ * l_adjustedPosZ * DRIVING_CONSTRAINT_SCALE_Z;
	
	l_force = etVector(l_forceX, l_forceY, l_forceZ);
	l_force.scale(DRIVING_STIFFNESS);

	return(l_force);
}


//***********************************************************
// calculate the all current haptics forces to a final force 
// value for this update frame
//***********************************************************
etVector haptics::calculateForces()
{
	// Tool position
    etVector l_posWC = m_terminusPositionServo;

	etVector l_force = etVector(0, 0, 0);

	int l_i;

	etVector l_relPos;
	double l_dotProd;
	etVector l_curForce;

	// update current surfaces
	for (l_i = 0; l_i < m_numSurfaces; l_i++)
	{
		// get the distance from the surface to the haptics pos
		
		// 1st, make the haptics pos relative to the surface pos
		l_relPos = l_posWC - m_surfaces[l_i].pos;

		// 2nd, get the dot product between the surface normal and the haptics vector
		l_dotProd = m_surfaces[l_i].normal.dot(l_relPos);

		if (l_dotProd < 0)
		{	// the dotProd is the projection of the haptics vector along the surface normal.
			// thus, it is the distance the haptics pos is from the surface
			l_curForce = m_surfaces[l_i].normal.scale(-l_dotProd);
			l_curForce = l_curForce.scale(m_surfaces[l_i].stiffness);

			l_force += l_curForce;
		}
	}

	// update current spheres
	for (l_i = 0; l_i < m_numSpheres; l_i++)
	{
		l_force += haptics::instance()->doSphereAt(m_spheres[l_i].pos, m_spheres[l_i].radius, l_posWC, m_spheres[l_i].stiffness);
	}
	
	// update current boxes
	for (l_i = 0; l_i < m_numBoxes; l_i++)
	{
		l_force += haptics::instance()->doBoxAt(m_boxes[l_i].pos, m_boxes[l_i].size, m_boxes[l_i].headVec, m_boxes[l_i].upVec, l_posWC, m_boxes[l_i].stiffness);
	}
	
	// update current forces
	for (l_i = 0; l_i < m_numForces; l_i++)
	{
		l_force += m_forces[l_i].dir.scale(m_forces[l_i].force);
	}

	//gravity force for holding general objects
    if (m_holdingWeight > 0)
    {
        l_force += etVector(0, -m_holdingWeight, 0);
    }

	if (m_bDoDrivingForces != 0)
	{
		l_force += haptics::doDrivingForces(l_posWC);
	}

	return(l_force);
}