//****************************************************************
// system for allocating entity objects
//****************************************************************


// object types
define TR1_OBJ_TYPE_NONE,					-1;	// undefined object
define TR1_OBJ_TYPE_CURSOR,				0;		// hand cursor
define TR1_OBJ_TYPE_PHYSICS_OBJ,			1;		// basic physics object
define TR1_OBJ_TYPE_PHYSICS_CAR,			2;		// physics driven vehicle
define TR1_OBJ_TYPE_SHRINK_BTN,			3;		// the shrink button
define TR1_OBJ_TYPE_GROW_BTN,				4;		// the grow button

define TR1_NUM_OBJ_TYPES,					5;		// number of object types

define TR1_NUM_OBJECTS,						20;	// number of objects in system

// haptics types
define TR1_HAPTICS_TYPE_NONE,		0;
define TR1_HAPTICS_TYPE_SPHERE,	1;
define TR1_HAPTICS_TYPE_BOX,		2;

// data values for each object (stored in skills)
// NOTE: the first 32 skills are reserved for the object system
define m_objType,				skill1;	// object type
define m_objID,				skill2;	// the object ID for this object

// collision and haptics settings for the object
define m_objPickupDist,		skill11;	// max distance from object that you can pick it up
define m_objNearDist,		skill12;	// max distance from object that will check for haptics effects

define m_hapticsWeight,		skill13;	// the haptics weight to apply when picking up the object

define m_hapticsType,		skill14;	// haptics type (none, sphere, box)
define m_hapticsStiffness,	skill15;	// haptics stiffness
define m_hapticsRadius,		skill16;	// radius of a haptics sphere
define m_hapticsBoxSize,	skill16;	// vector (3 vars) storing the extents of a haptics box


// object buffer
var gtr1_ObjectHandle[TR1_NUM_OBJECTS];
var gtr1_ObjectNext[TR1_NUM_OBJECTS];

// unused object list top
var gtr1_UnusedObjs;

// active object list tops
var gtr1_ActiveObjs[TR1_NUM_OBJ_TYPES];

// other vars
entity* gtr1_pCurIDEntity;
entity* gtr1_pAltIDEntity;

// input varaible for tr1_Object_SetHapticsBox
var gtr1_BoxSize[3];



//********************************************************
// init the object system
//********************************************************
function tr1_Object_InitSystem ()
{
	var l_i;
	
	// link up the list of objects
	l_i = 0;
	while (l_i < (TR1_NUM_OBJECTS-1))
	{
		gtr1_ObjectHandle[l_i] = -1;
		gtr1_ObjectNext[l_i] = l_i + 1;
		
		l_i += 1;
	}
	
	// end the list
	gtr1_ObjectHandle[l_i] = -1;
	gtr1_ObjectNext = -1;
	
	// set up the unused list to the top of the buffer
	gtr1_UnusedObjs = 1;
	
	// set the active lists to empty
	l_i = 0;
	while (l_i < TR1_NUM_OBJ_TYPES)
	{
		gtr1_ActiveObjs[l_i] = -1;
		
		l_i += 1;
	}
}


//********************************************************
// grab a new unused object, adding it to its appropriate 
// active list
//
// IN:  type of object
//
// RETURNS:  ID (index) of the object (-1 if none left)
//
// NOTE: the procedure for creating a new object is:
//				1) call objID = tr1_Object_New(objType)
//				2) create the new enity (using ent_create)
//				3) get the handle for the new entity (using handle())
//				4) call tr1_Object_Set(objID, handle, objType)
//********************************************************
function tr1_Object_New (v_objType)
{
	var l_newObjID;
	
	if (gtr1_UnusedObjs < 0)
	{	// no objects left
		return(-1);
	}
	
	if ((v_objType <= TR1_OBJ_TYPE_NONE) || (v_objType >= TR1_NUM_OBJ_TYPES))
	{	// object type out of range
		return(-1);
	}
	
	// pop the top object off the unused list
	l_newObjID = gtr1_UnusedObjs;
	gtr1_UnusedObjs = gtr1_ObjectNext[l_newObjID];
	
	// add the object to the top of the appropriate active list
	gtr1_ObjectNext[l_newObjID] = gtr1_ActiveObjs[v_objType];
	gtr1_ActiveObjs[v_objType] = l_newObjID;
	
	return(l_newObjID);
}


//********************************************************
// delete an active object
//
// IN:  object ID
//********************************************************
function tr1_Object_Delete (v_objID)
{
	var l_objHandle;
	var l_objType;
	var l_objNext;
	
	var l_curID;
	
	if ((v_objID < 0) || (v_objID >= TR1_NUM_OBJECTS))
	{	// id is out of range
		return;
	}
	
	// grab the handle for the object
	l_objHandle = gtr1_ObjectHandle[v_objID];
	
	if (l_objHandle == -1)
	{	// no entity handle registered for this id
		return;
	}
	
	// grab the entity pointer
	gtr1_pCurIDEntity = ptr_for_handle(l_objHandle);
	if (gtr1_pCurIDEntity == NULL)
	{	// illegal handle
		return;
	}
	
	// grab the object type
	l_objType = gtr1_pCurIDEntity.m_objType;
	
	if (v_objID == gtr1_ActiveObjs[l_objType])
	{	// this object is at the top of its active list
	
		// remove it from its active list
		gtr1_ActiveObjs[l_objType] = gtr1_ObjectNext[v_objID];
		
		// add it to the unused list
		gtr1_ObjectNext[v_objID] = gtr1_UnusedObjs;
		gtr1_UnusedObjs = v_objID;

		gtr1_ObjectHandle[v_objID] = -1;
	}
	else
	{	// we need to search for it
		l_curID = gtr1_ActiveObjs[l_objType];
		
		while (gtr1_ObjectNext[l_curID] >= 0)
		{
			if (gtr1_ObjectNext[l_curID] == v_objID)
			{	// found it... remove it from the active list
				gtr1_ObjectNext[l_curID] = gtr1_ObjectNext[v_objID];
				
				// add it to the unused list
				gtr1_ObjectNext[v_objID] = gtr1_UnusedObjs;
				gtr1_UnusedObjs = v_objID;
				
				gtr1_ObjectHandle[v_objID] = -1;
				
				break;
			}
		}
	}
}


//********************************************************
// set the values for a new object
//
// IN:	v_objID = the ID (index) of the object
//			v_handle = the entity handle of the object
//			v_objType = the type of object
//********************************************************
function tr1_Object_Set (v_objID, v_handle, v_objType)
{
	if ((v_objID < 0) || (v_objID >= TR1_NUM_OBJECTS))
	{	// id is out of range
		return;
	}
	
	if ((v_objType <= TR1_OBJ_TYPE_NONE) || (v_objType >= TR1_NUM_OBJ_TYPES))
	{	// object type out of range
		return(-1);
	}

	// grab the entity pointer
	gtr1_pCurIDEntity = ptr_for_handle(v_Handle);
	if (gtr1_pCurIDEntity == NULL)
	{	// illegal handle
		return;
	}

	// set the entity handle to the object
	gtr1_ObjectHandle[v_objID] = v_handle;
	
	// set the object type and ID in the entity skills
	gtr1_pCurIDEntity.m_objType = v_objType;
	gtr1_pCurIDEntity.m_objID = v_objID;
	
	// setup default settings for collision/haptics
	gtr1_pCurIDEntity.m_objPickUpDist = 100.0;
	gtr1_pCurIDEntity.m_objNearDist = 200.0;
	gtr1_pCurIDEntity.m_hapticsWeight = 0.0;
	gtr1_pCurIDEntity.m_hapticsType = TR1_HAPTICS_TYPE_NONE;
}


//********************************************************
// return the entity handle for an object
//
// IN:  object ID
//
// RETURNS: entity handle
//********************************************************
function tr1_Object_GetHandle (v_objID)
{
	return(gtr1_ObjectHandle[v_objID]);
}


//********************************************************
// return the entity ptr for an object
//
// IN:  object ID
//
// OUT: gtr1_pCurIDEntity holds the object's entity ptr
//********************************************************
function tr1_Object_GetEntity (v_objID)
{
	gtr1_pCurIDEntity = ptr_for_handle(gtr1_ObjectHandle[v_objID]);
}


//********************************************************
// return the object type for an object
//
// IN:  object ID
//
// RETURNS: object's type
//
// NOTE: sets gtr1_pCurIDEntity to the object's entity ptr
//********************************************************
function tr1_Object_GetType (v_objID)
{
	gtr1_pCurIDEntity = ptr_for_handle(gtr1_ObjectHandle[v_objID]);
	
	return(gtr1_pCurIDEntity.m_objType);
}


//********************************************************
// set the basic collide/haptics settings for an object
//
// IN: v_objID = object ID
//     v_pickupDist = the distance at which the cursor can pick up the object
//     v_nearDist = the distance at which haptics checking begins on this object
//     v_hapticsWeight = the haptics weight to apply when picked up
//********************************************************
function tr1_Object_SetCollideInfo (v_objID, v_pickupDist, v_nearDist, v_hapticsWeight)
{
	gtr1_pCurIDEntity = ptr_for_handle(gtr1_ObjectHandle[v_objID]);

	gtr1_pCurIDEntity.m_objPickupDist = v_pickupDist;
	gtr1_pCurIDEntity.m_objNearDist = v_nearDist;
	gtr1_pCurIDEntity.m_hapticsWeight = v_hapticsWeight;
}


//********************************************************
// set an object to have a haptics sphere
//
// IN: v_objID = object ID
//     v_radius = radius of the haptics sphere
//		 v_stiffness = the stiffness of the sphere
//********************************************************
function tr1_Object_SetHapticsSphere (v_objID, v_radius, v_stiffness)
{
	gtr1_pCurIDEntity = ptr_for_handle(gtr1_ObjectHandle[v_objID]);
	
	gtr1_pCurIDEntity.m_hapticsType = TR1_HAPTICS_TYPE_SPHERE;
	gtr1_pCurIDEntity.m_hapticsRadius = v_radius;
	gtr1_pCurIDEntity.m_hapticsStiffness = v_stiffness;
}


//********************************************************
// set an object to have a haptics box
//
// IN: v_objID = object ID
//		 v_stiffness = the stiffness of the box
//     gtr1_BoxSize = vector holding the extents of the box
//********************************************************
function tr1_Object_SetHapticsBox (v_objID, v_stiffness)
{
	gtr1_pCurIDEntity = ptr_for_handle(gtr1_ObjectHandle[v_objID]);
	
	gtr1_pCurIDEntity.m_hapticsType = TR1_HAPTICS_TYPE_BOX;
	gtr1_pCurIDEntity.m_hapticsStiffness = v_stiffness;
	vec_set(gtr1_pCurIDEntity.m_hapticsBoxSize, gtr1_BoxSize);
}


