//****************************************************************
//* basic physics system functions, and generic object physics
//****************************************************************


var TR1_OBJ_HOLDING_VELOCITY_SCALE;		// velocity scale to use when moving the object around


// object local vars, stored in the entity's skills
// NOTE: skill1-skill32 are reserved for the object system
define m_mass,			skill33;		// mass of this object
define m_friction,	skill34;		// friction on the object
define m_damping,		skill35;		// damping on this object

// inputs to Physics_<ObjType>_Create
var gtr1_ObjCreatePos[3];
var gtr1_ObjCreateAng[3];


//***************************************************************
// setup the global physics values
//***************************************************************
function tr1_Physics_GlobalInit ()
{
	ph_setgravity(vector(0, 0, -tr1_dll_getPhysicsGravity()));
	
	// load .ini file constants
	TR1_OBJ_HOLDING_VELOCITY_SCALE = tr1_dll_getObjHoldingVelocityScale();
}


//****************************************************************
// behavior routine for generic physics objects
//****************************************************************
action tr1_aPhysicsObj_Update ()
{
	var l_bPhysicsEnabled = 0;
	
	var l_vecToCursor[3];
	var l_holdPos[3];
	
	// wait for skill vars to be setup
	wait(1);

	my.fat = on; 
	my.narrow = on; 
	c_setminmax(my);

	my.shadow = on;

	my.passable = on;
	
	// setup physics body for the object
	phent_settype(my, PH_RIGID, PH_BOX);
	phent_setmass(my, my.m_mass, PH_BOX);
	phent_setgroup(my, 1);
	phent_setfriction(my, my.m_friction);
	phent_setdamping(my, my.m_damping, my.m_damping);
	phent_setelasticity(my, 0, 0); // little bouncing
				
	my.passable = off;

	// set the orientation of the obj
	vec_set(my.pan, gtr1_ObjCreateAng);
	
	// update loop
	while (1)
	{
		if (my != gtr1_pObjHeld)
		{	// I'm not being held by the cursor
			if (l_bPhysicsEnabled == 0)
			{	
				// set for dropping
				phent_setdamping(my, my.m_damping, my.m_damping);
				
				l_bPhysicsEnabled = 1;

				// set my push values to be higher than the cursor
				my.push = TR1_OBJECT_PUSH;
			}
		}
		else
		{	// currently being held by the cursor
			if (l_bPhysicsEnabled != 0)
			{	
				phent_setdamping(my, 100, 100);
				
				l_bPhysicsEnabled = 0;

				// set my push values to be lower than the cursor
				my.push = TR1_HELDOBJ_PUSH;
			}
			
			// halt previous velocity
			phent_clearvelocity(my);
			
			if (gtr1_CameraMode == TR1_CAMERA_MODE_PICKUP)
			{	// use the cursor's current position
				tr1_Object_GetEntity(gtr1_CursorID);
				vec_set(l_holdPos, gtr1_pCurIDEntity.x);
			}
			else
			{	// use the cursor's last position, relative to camera
				vec_set(l_holdPos, gtr1_CursorPos);
				vec_rotate(l_holdPos, camera.pan);
				vec_add(l_holdPos, camera.x);
			}
					
			// get velocity to move by
			vec_diff(l_vecToCursor, l_holdPos, my.x);
			vec_scale(l_vecToCursor, TR1_OBJ_HOLDING_VELOCITY_SCALE);
			
			phent_addvelcentral(my, l_vecToCursor);
		}

		wait(1);
	}
}

	
//---------------------------------------------------------
// FLOWERPOT object
//---------------------------------------------------------


//*******************************************************
// create a flowerpot physics object
//
// IN:  gObjCreatePos = vector holding the position to launch the obj at
//      gObjCreateAng = the rotation of the obj
//
// RETURNS: object ID of the flowerpot
//*******************************************************
function tr1_Physics_Flowerpot_Create ()
{
	var l_objID;
	var l_objHandle;
	var l_scale;

	l_objID = tr1_Object_New(TR1_OBJ_TYPE_PHYSICS_OBJ);
	if (l_objID < 0)
	{	// error... no objects left
		return(-1);
	}
	
	// create the flowerpot
	gtr1_pCurIDEntity = ent_create("emptypot.mdl", gtr1_ObjCreatePos, tr1_aPhysicsObj_Update);
	l_objHandle = handle(gtr1_pCurIDEntity);
	
	// set its orientation
	vec_set(gtr1_pCurIDEntity.pan, gtr1_ObjCreateAng);

	// set the object settings
	tr1_Object_Set(l_objID, l_objHandle, TR1_OBJ_TYPE_PHYSICS_OBJ);
	
	// set the scale
	l_scale = tr1_dll_getFlowerpotScale();
	gtr1_pCurIDEntity.scale_x = l_scale;
	gtr1_pCurIDEntity.scale_y = l_scale;
	gtr1_pCurIDEntity.scale_z = l_scale;

	// add all the info for this type of object
	gtr1_pCurIDEntity.m_mass = tr1_dll_getFlowerpotPhysicsMass();
	gtr1_pCurIDEntity.m_friction = tr1_dll_getFlowerpotPhysicsFriction();
	gtr1_pCurIDEntity.m_damping = tr1_dll_getFlowerpotPhysicsDamping();
	
	// setup collision/haptics for the car
	tr1_Object_SetCollideInfo(l_objID, tr1_dll_getFlowerpotCollidePickupSize(), tr1_dll_getFlowerpotCollideNearSize(), tr1_dll_getFlowerpotHapticsWeight());
	tr1_Object_SetHapticsSphere(l_objID, tr1_dll_getFlowerpotHapticsRadius(), tr1_dll_getSphereStiffness());

	return(l_objID);
}


//---------------------------------------------------------
// CRATE object
//---------------------------------------------------------


//*******************************************************
// create a crate physics object
//
// IN:  gObjCreatePos = vector holding the position to launch the obj at
//      gObjCreateAng = the rotation of the obj
//
// RETURNS: object ID of the flowerpot
//*******************************************************
function tr1_Physics_Crate_Create ()
{
	var l_objID;
	var l_objHandle;
	var l_scale;

	l_objID = tr1_Object_New(TR1_OBJ_TYPE_PHYSICS_OBJ);
	if (l_objID < 0)
	{	// error... no objects left
		return(-1);
	}
	
	// create the crate
	gtr1_pCurIDEntity = ent_create("crate.mdl", gtr1_ObjCreatePos, tr1_aPhysicsObj_Update);
	l_objHandle = handle(gtr1_pCurIDEntity);
	
	// set its orientation
	vec_set(gtr1_pCurIDEntity.pan, gtr1_ObjCreateAng);

	// set the object settings
	tr1_Object_Set(l_objID, l_objHandle, TR1_OBJ_TYPE_PHYSICS_OBJ);
	
	// set its scale
	l_scale = tr1_dll_getCrateScale();
	gtr1_pCurIDEntity.scale_x = l_scale;
	gtr1_pCurIDEntity.scale_y = l_scale;
	gtr1_pCurIDEntity.scale_z = l_scale;

	// add all the info for this type of object
	gtr1_pCurIDEntity.m_mass = tr1_dll_getCratePhysicsMass();
	gtr1_pCurIDEntity.m_friction = tr1_dll_getCratePhysicsFriction();
	gtr1_pCurIDEntity.m_damping = tr1_dll_getCratePhysicsDamping();

	// setup collision/haptics for the car
	tr1_Object_SetCollideInfo(l_objID, tr1_dll_getCrateCollidePickupSize(), tr1_dll_getCrateCollideNearSize(), tr1_dll_getCrateHapticsWeight());
	
	vec_set(gtr1_BoxSize, vector(tr1_dll_getCrateHapticsSizeX(), tr1_dll_getCrateHapticsSizeY(), tr1_dll_getCrateHapticsSizeZ()));
	tr1_Object_SetHapticsBox(l_objID, tr1_dll_getBoxStiffness());

	return(l_objID);
}
